(* 
 * Copyright (c) 2001, 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: gontutil.ml,v 1.17 2002/06/04 11:54:02 malekith Exp $
 *
 * Module interface: Gontutil
 * Description: Various utilities definitions for Gont compiler.
 *)

open Gonttree
;;

exception Syntax_error of location * string
exception Error of string
exception Eof

;;
(* Utils. *)
let warnings_count = ref 0

let xf = Printf.sprintf

let err_msg (loc, msg) =
  prerr_string (xf "%s:%d:%d: %s" 
                   loc.loc_file loc.loc_line (loc.loc_char+1)
                   msg);
  prerr_newline ()

let warn (loc, msg) = 
  incr warnings_count;
  err_msg (loc, ("warning: " ^ msg))

let err (loc, msg) = raise (Syntax_error(loc, msg))

let debug s = print_string s; print_newline ()

let rec list_apply fnc (c, xxs) =
  match xxs with
      x::xs -> 
        let (c', x') = fnc (c, x) in
        let (c'', xs') = list_apply fnc (c', xs) in
        (c'', x'::xs')
    | [] -> (c, [])

let sequences : (string, int ref) Hashtbl.t = Hashtbl.create 16

let reset () = Hashtbl.clear sequences

let seq_name s =
  let r = 
    (try
       Hashtbl.find sequences s 
     with Not_found -> 
       let n = ref 0 in
       Hashtbl.add sequences s n; n) in
  incr r; s ^ (string_of_int !r)

let rec type_walk f t =
  let fs = type_walk f in
  let fl = List.map fs in
  let ff = List.map (fun ({fr_type = t} as r) -> {r with fr_type = fs t}) in
  let t' = f t in
  match t' with
      Int | Void | Bool | Float | String | TyVar _ | AbstractType | Exn -> t'
    | Tuple l -> Tuple (fl l)
    | Func {ft_ret_type = r; ft_parms = p; ft_tyvars = t} ->
       Func {ft_ret_type = fs r; ft_parms = fl p; ft_tyvars = t}
    | Struct l -> Struct  (ff l)
    | OptStruct l -> OptStruct (ff l)
    | Union l -> Union (ff l)
    | NamedType ({nt_parms = p} as r) -> 
       NamedType {r with nt_parms = fl p}

let simple_fv t =
  let res : id list ref = ref [] in
  let f t =
    match t with
        TyVar(a) ->
          if not (List.mem a !res) then
            res := a :: !res;
          t
      | NamedType _ | Int | Void | Bool | Float | String | AbstractType
      | Exn | Tuple _ | Func _ | Struct _ | OptStruct _ | Union _ -> t in
  ignore(type_walk f t);
  !res

let make_fn_type (r, a) =
  let a' = List.map (fun {parm_type = t} -> t) a in
  let t = {ft_ret_type = r; ft_parms = a'; ft_tyvars = []} in
  let fv = simple_fv (Func t) in
  Func {t with ft_tyvars = fv}
;;

let has_property sym p = List.mem p sym.sym_properties
let set_property sym p = 
  if has_property sym p then
    ()
  else
    sym.sym_properties <- p :: sym.sym_properties
let union_typedef sym =
  match sym with
  | {sym_typedefs = [t]} when (has_property sym IsUnionField) -> t
  | _ -> assert false
let union_field_id sym =
  let rec fnd l =
    match l with
    | (UnionFieldId k)::xs -> k
    | _::xs -> fnd xs
    | [] -> assert false in
  fnd sym.sym_properties
let sym_module s =
  try
    String.sub s.sym_name 0 ((String.rindex s.sym_name ':') - 1)
  with Not_found -> raise (Invalid_argument "sym_module")
let sym_eq s1 s2 =
  assert (s1.sym_kind = s2.sym_kind);  
  s1.sym_name = s2.sym_name
;;
